# ------------------------------ MNI Header ----------------------------------
#@NAME       : JobControl.pm
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: A module for doing process and job stuff -- spawning 
#              children and waiting for them (or detaching them to the
#              background), detaching ourself to the background, running
#              stuff remotely -- that sort of thing.
#@METHOD     : 
#@GLOBALS    : There are a number of options that are global variables
#              within the module, but shouldn't really be thought of
#              that way by the outside world.  These should all be 
#              set via the SetOptions subroutine, e.g.
#              
#                 &JobControl::SetOptions ("Verbose", 1);
#              
#              to be verbose.  The currently valid options are:
#              
#                  Verbose     print commands as we execute them?
#		   Execute     actually execute commands?
#                  MergeErrors merge stderr with stdout?
#                  ErrorAction what to do when a command fails
#                  Batch       submit commands to batch system?
#                  Clobber     overwrite output files (instead of append)
#                  LogHandle   filehandle to write commands to
#                  Notify      should &Obituary actually send mail?
#
#              these are all further documented with the routines that
#              actually use them (eg. see &Spawn for the gory details on
#              MergeErrors, &Backgroundify and/or &Spawn for info on
#              LogHandle, etc.
#@CALLS      : 
#@CREATED    : 1995/03/03, Greg Ward
#@MODIFIED   : 1995/11/14, GW: made into a package
#              1996/04/11, GW: made into a Perl 5 module
#@VERSION    : $Id: JobControl.pm,v 1.27 1996/10/03 17:55:40 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

require 5.001;
require "misc_utilities.pl"; # for &userstamp and &timestamp

package JobControl;

use FileHandle;			# for autoflush method
use Carp;

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(SetSpawnOptions Backgroundify Spawn Obituary);


$Notify = $ENV{'USER'};
$ErrorAction = "";
$MergeErrors = 1;
$Batch = 0;
$Clobber = 0;
$LogHandle = "STDOUT";

%valid_options = ("Verbose"     => 1, # print commands as we execute them?
		  "Execute"     => 1, # actually execute commands?
		  "MergeErrors" => 1, # merge stderr with stdout?
		  "ErrorAction" => 1, # what to do when a command fails
		  "Batch"       => 1, # submit commands to batch system?
		  "Clobber"     => 1, # overwrite output files (not append)
		  "LogHandle"   => 1, # filehandle to write commands to
		  "Notify"      => 1); # should &Obituary actually send mail?
		  

sub SetOptions
{
   local (@optval) = @_;

   ($filename,$line) = (caller)[1..2]; # for error messages

   if (@optval == 2 && $optval[0] =~ /\d+/)
   {
      warn "warning: obsolete version of \&SetSpawnOptions called from " .
	 "${filename}:${line}\n";

      ($MergeErrors, $ErrorAction) = @optval;
      return;
   }

   while (@optval)
   {
      $opt = shift @optval;
      $val = shift @optval;
      ($opt = "Clobber", warn "ClobberLogs option is obsolete - use Clobber\n")
         if $opt eq "ClobberLogs";
      croak ("Unmatched option $opt at ${filename}:${line}")
	 unless defined $val;
	 
      croak ("Unknown option $opt at ${filename}:${line}") unless 
	 $valid_options{$opt};

      eval "\$$opt = \$val";
   }
}


sub SetSpawnOptions
{
   &SetOptions;
}


sub set_undefined_flags
{
   my (@flags) = @_;

   foreach $flag (@flags)
   {
      eval "\$$flag = \$::$flag unless defined \$$flag";
   }
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : Backgroundify
#@INPUT      : $logfile - name of file to which the stdout and stderr should
#                         be redirected (if empty, will try to use LogHandle
#                         package option)
#              @args    - arguments program was run with (used to generate
#                         announcement message) (just a copy of @ARGV)
#                         (if not supplied, won't print out banner describing
#                         how program was run)
#@OUTPUT     : none
#@RETURNS    : 0 if redirection of stdout or stderr failed (does NOT fork)
#              0 if caller didn't specify anywhere to redirect to, either
#                 with $logfile parameter or LogHandle package option
#              if successful: returns 1 to newly-detached child process
#                             never returns to parent (it terminates!)
#@DESCRIPTION: Redirects stdout and stderr to a file (named by $logfile),
#              and detaches to background execution.  If @args is non-
#              empty, prints two announcements: to stdout, that we're
#              detaching to the background; and to the logfile, the
#              user, host, time, and full argument list (@args).
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/02/21, Greg Ward: copied from nearly-identical function
#                 in do_mritopet
#@MODIFIED   : 1996/03/15-26, GW: changed to pay attention to Clobber 
#                 and LogHandle options
#@COMMENTS   : 
#-----------------------------------------------------------------------------
sub Backgroundify
{
   my ($logfile, @args) = @_;
   my ($time, $user, $log_existed, $stdout);

   &set_undefined_flags ("Verbose");

   # If caller didn't supply a log filename, check the package option
   # LogHandle -- if it's anything other than STDOUT, then we'll 
   # redirect to it.

   if (! $logfile)
   {
      if ($LogHandle ne "STDOUT")
      {
	 print "Redirecting output and detaching to background\n"
	    if $Verbose;
	 $stdout = ">&$LogHandle";
      }
      else
      {
	 warn "Backgroundify: must specify either \$logfile or LogHandle\n";
	 return 0;
      }
   }

   # Caller specified a filename, so use it and ignore LogHandle option.

   else
   {
      print "Redirecting output to $logfile and detaching to background\n"
	 if $Verbose;
      $stdout = ($Clobber ? ">" : ">>") . "$logfile";
   }

   unless (open (STDOUT, $stdout))
   {
      warn "Detachment to background failed: couldn't redirect stdout ($!)\n";
      return 0;
   }
   unless (open (STDERR, ">&STDOUT"))
   {
      warn "Detachment to background failed: couldn't redirect stderr ($!)\n";
      return 0;
   }
	   
   autoflush STDERR 1;
   autoflush STDOUT 1;

   if (@args)
   {
      print "\n" . "=" x 80 . "\n"
	 if ($log_existed && !$Clobber);
      printf "%s %s opening log file for:\n", &::userstamp(), &::timestamp();
      print "$0 " . join(" ", @ARGV) . "\n\n";
   }

   fork && exit;
   return 1;			# to the *child* -- parent terminates!
}




# ------------------------------ MNI Header ----------------------------------
#@NAME       : &check_return_status
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Checks the return status of a program.  If non-zero,
#              prints a warning and eval's $ErrorAction (if it
#              is defined).
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 96/01/17, GPW (from code in &Spawn)
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub check_return_status
{
   my ($status, $program, $command) = @_;

   if ($status)
   {
      if ($ErrorAction eq "fatal")
      {
         &::Fatal ("crashed while running $program (exit status=$status)");
      }
      elsif ($ErrorAction eq "notify")
      {
         &Obituary ($command, $status);
      }
      elsif ($ErrorAction)
      {
	 my ($i, $this_pkg, @caller);

	 $i = 0;
	 $this_pkg = (caller(0))[0];
	 while (@caller = caller($i++))
	 {
	    $up_pkg = $caller[0];
	    last if $up_pkg ne $this_pkg;
	 }

	 eval "package $up_pkg; $ErrorAction";
	 croak ("Error in eval'd code: $@\n") if $@;
         return 0;
      }
      else 
      {
         warn "$program crashed (exit status=$status)\n";
         return 0;
      }
   }
}



# ------------------------------ MNI Header ----------------------------------
#@NAME       : &Spawn
#@INPUT      : $command - the command (and all arguments) to run
#              $stdout  - tells what to do with the standard output.
#                         Can be either a filename (redirect to that
#                         file), "-" (capture to a variable, which is
#                         returned by &Spawn), or empty (leave it alone).
#                         If $stdout is a filename, it will be either
#                         overwritten or appended depending on the 
#                         Clobber option.
#              $stderr  - identical to $stdout.  Note that you
#                         can merge the standard error into the standard
#                         output by using ">&STDOUT" for $stderr (this is
#                         a standard Perl trick).
#              $stdin   - doesn't do anything (yet) (and probably never will)
#@OUTPUT     : 
#@RETURNS    : $status - exit status of the child process
#              $output - standard output of the child process (as long as
#                        $stdout was empty or "-")
#              $error  - standard error of the child process (as long as
#                        $stderr was empty or "-")
#@DESCRIPTION: Runs a child process, optionally capturing its standard
#              output and/or standard error to a file or a Perl variable.
#              To capture either stdout or stderr to a filename, just pass
#              filenames (eg. "foo.log") as $stdout and/or $stderr.  You
#              can merge stderr with stdout either by setting $stderr to
#              ">&STDOUT", or by leaving $stderr empty (or undefined) and
#              setting the MergeErrors option to a true value.
#              
#              To capture to a variable, simply evaluate &Spawn in an array
#              context, eg. ($status, $output, $error) = &Spawn (...).  By
#              default, this will make stdout and stderr go to separate
#              variables; you can counteract this (as above) by setting
#              $stderr to ">&STDOUT" or by setting $MergeErrors to 2.
#@METHOD     : 
#@GLOBALS    : $Verbose, $Execute, $MergeErrors, $ErrorAction
#@CALLS      : 
#@CREATED    : 95/3/3, Greg Ward (didn't really get anywhere)
#@MODIFIED   : 95/5/16, GW - finished it (well, mostly)
#              95/6/20, GW - added code for capturing stdout and stderr
#-----------------------------------------------------------------------------
sub Spawn
{
   local ($command, $stdout, $stderr, $stdin) = @_;
   local ($program);
   local ($pid, $status, $output, $error); # must be dynamically scoped!!!

   &set_undefined_flags ("Verbose", "Execute");

   $stdout = "" if (! defined $stdout);
   $stderr = "" if (! defined $stderr);
   $stdin = "" if (! defined $stdin);

   # Check for unimplemented features

   croak ("&Spawn: sorry, can't redirect stdin at all (yet)\n")
      if ($stdin ne "");

   # If Batch mode is on, submit this command and be done with it

   if ($Batch && $Execute)
   {
      if (! defined &Batch::QueueCommand)
      {
         croak "Batch package not loaded -- you must do \"use Batch\"";
      }

      $stdout = "" if $stdout eq "-"; # can't capture
      $stderr = "" if $stderr eq "-";
      &Batch::QueueCommand ($command, "", $stdout, $stderr, $MergeErrors);
      return 0;
   }

   # Get the first word (i.e. the program to run) from the command string

   ($program) = split (/\s+/, $command);

   # Capture both stdout and stderr to variables if we're evaluated
   # in an array context and a redirection destination (ie. file) 
   # hasn't already been specified for them.

   $stdout = "-" if ($stdout eq "" && wantarray);
   $stderr = "-" if ($stderr eq "" && wantarray);

   # If nothing was explicitly said about what to do with $stderr,
   # follow the advice of $MergeErrors (i.e. if capturing to a variable,
   # then merge stderr only if $MergeErrors is 2; if not capturing, then
   # merge by default if $MergeErrors is 1).

   if ($stderr eq "")
   {
      $stderr = ">&STDOUT" if $MergeErrors >= 1;
   }
   elsif ($stderr eq "-")
   {
      $stderr = ">&STDOUT" if $MergeErrors >= 2;
   }

   # Figure out how to open files (> to overwrite or >> to append),
   # and hard-code that into $stdout and $stderr if they don't
   # already have such a code in them.

   $open_prefix = ($Clobber) ? ">" : ">>";
   $stdout = $open_prefix . $stdout if ($stdout && $stdout !~ /^[>-]/);
   $stderr = $open_prefix . $stderr if ($stderr && $stderr !~ /^[>-]/);

   # Report the user, host, date, time, and command to be run

   printf $LogHandle "%s %s %s\n", &::userstamp(), &::timestamp(), $command
      if $Verbose;
   return 0 unless $Execute;

   # Do things rather differently depending on whether we are to
   # capture stdout.  In the capture-to-variable case, we will
   # fork-and-exec the desired command, and slurp its output via a
   # pipe; stderr will be captured via a temporary file (if stderr is
   # supposed to be captured - it might be redirected or left alone).
   # (Yeah, I know stdout could just as well have been captured the
   # same way.  However, I was 1) curious about doing it via a pipe,
   # and 2) concerned about efficiency -- error output is usually
   # pretty small so going via disk won't kill anyone, but stdout
   # could be big.)  I didn't do stderr via a pipe because I don't
   # think you can have *two* pipes (one for stdout and one for
   # stderr) connecting the two processes.  Perhaps it could be done
   # with a named pipe... ah well, another day...

   if ($stdout eq "-")		# capturing to a variable
   { 
      # Run the command through a pipe.  In the child, stdout
      # automagically goes to the pipe; in the parent, we will slurp
      # the child's output from the pipe.

      $pid = open (PIPE, "-|");
      croak ("&Spawn: failed to start child process: $!\n") 
         unless defined $pid;
      if ($pid == 0)		# in the child?  
      {
	 # We're in the child so don't do much: do error redirection if
	 # any (including merging into stdout - this is handled above),
	 # and then exec.

	 if ($stderr eq "-")	# caller wants it in a variable -- we'll
	 {			# be sneaky and put it in a file for later
	    open (STDERR, ">/tmp/error$$.log");	# retrieval 
	 }
	 elsif ($stderr ne "")	# caller wants it in a file (including
	 {			# possibly &STDOUT), so just put it there
	    open (STDERR, $stderr);
	 }

	 if (! exec ($command))
	 {
	    croak ("Unable to execute $program: $!\n");
	 }
      }
      else			# $pid != 0 => in the parent
      {
	 # We're in the parent, so we read in the child's stdout
	 # (and stderr too if it was captured).

	 $output = join ("", <PIPE>);
	 close (PIPE);
	 $status = $? >> 8;

	 if ($stderr eq "-")	# was stderr captured?
	 {
	    open (ERROR, "</tmp/error${pid}.log");
	    $error = join ("", <ERROR>);
	    close (ERROR);
	    unlink ("/tmp/error${pid}.log");
	 }

	 &check_return_status ($status, $program, $command);
         return ($status, $output, $error) if wantarray;
         return $status;
      }
   }

   # If capturing to file(s), do it right -- fork, do redirection 
   # in the child process, and exec the command; in the parent, just
   # wait for the child to finish

   $pid = fork;
   croak ("&Spawn: failed to start child process: $!\n") unless defined $pid;
   if ($pid == 0)		# we're in the child process
   {
      # Redirect stdout and/or stderr if the caller specified file 
      # destination(s)

      if ($stdout ne "") 
      {
	 open (STDOUT, $stdout) 
	    || croak ("Unable to redirect stdout to $stdout: $!\n");
      }
      if ($stderr eq "-") 
         { open (STDERR, ">/tmp/error$$.log"); }
      elsif ($stderr ne "")
         { open (STDERR, $stderr); }

      if (! exec ($command))
      {
	 croak ("Unable to execute $program: $!\n");
      }
   }
   else				# we're still in the parent
   {
      # block and wait for child

      if (waitpid ($pid, 0) == -1)
      {
	 confess ("&Spawn: no child processes even though I just forked!");
      }
      
      $status = $? >> 8;

      if ($stderr eq "-")	# was stderr captured?
      {
	 open (ERROR, "</tmp/error${pid}.log");
	 $error = join ("", <ERROR>);
	 close (ERROR);
	 unlink ("/tmp/error${pid}.log");
      }

      &check_return_status ($status, $program, $command);

      return ($status, $output, $error) if wantarray;
      return $status;
   }      
}


sub Obituary                    # use this for long-running programs that
{				# send mail when done or crashed
   my ($command, $status, $error) = @_;
   my ($program, $text);

   &set_undefined_flags ("Notify", "Execute");

   if ($Notify && $Execute)
   {
      if ($command)
      {
	 $program = (split (/\s+/, $command))[0];
	 $text = "$::ProgramName crashed while executing the command:\n";
         $text .= $command;
      }
      else
      {
	 $text = "$::ProgramName crashed";
      }
      open (MAIL, "|/usr/lib/sendmail $Notify");
      
      print MAIL <<EOM;
From: $::ProgramName
Subject: $::ProgramName crashed!
	 
Dear $Notify,
	 
$text
	    
EOM
      print MAIL "Exit status: $status\n" if ($status);
      if (defined $error)
      {
	 print MAIL "Error message(s): \n$error\n";
      }
      else
      {
	 print MAIL "You'll have to check the log file to see what exactly went wrong.\n";
      }

      close (MAIL);
   }

   if ($program)
      { &::Fatal ("crashed while running $program"); }
   else
      { &::Fatal ("crashed"); }
}



1;
