# ------------------------------ MNI Header ----------------------------------
#@NAME       : file_utilities.pl
#@DESCRIPTION: Functions to manipulate/check/validate/search files and
#              directories.
#@GLOBALS    : 
#@CREATED    : 1995/08/07, Greg Ward (from gpw_utilities.pl)
#@MODIFIED   : 
#@VERSION    : $Id: file_utilities.pl,v 1.12 1996/08/01 13:13:10 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

require 5.001;

# Aliases for the old way of naming these subs:

sub CheckOutputDirs { &check_output_dirs; }
sub CheckOutputPath { &check_output_path; }
sub CheckInputDirs { &check_input_dirs; }
sub CheckDirectories { &check_directories; }
sub SearchDirectories { &search_directories; }
sub CheckFiles { &check_files; }
sub FindProgram { &find_program; }
sub FindPrograms { &find_programs; }
sub Uncompress { &uncompress; }



# ------------------------------ MNI Header ----------------------------------
#@NAME       : check_output_dirs
#@INPUT      : @dirs - list of directories to check
#@OUTPUT     : 
#@RETURNS    : 1 if all directories exist and are writeable, *or* were
#                 successfully created
#              0 otherwise
#@DESCRIPTION: Checks that each directory in a list of directories either
#              exists and is writeable; *or* that it can be created.  
#              Prints a meaningful error message and returns 0 if any fail.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/05/02, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub check_output_dirs
{
   local (@dirs) = @_;
   local ($num_err) = 0;

   foreach $dir (@dirs)
   {
      next unless $dir;		# skip blank strings
      if (-e $dir)
      {
	 if (! -d $dir)
	 {
	    warn "$dir already exists but is not a directory\n";
	    $num_err++;
	 }
	 elsif (! -w $dir)
	 {
	    warn "$dir is a directory but is not writeable\n";
	    $num_err++;
	 }
      }
      else
      {
	 if (! mkdir ($dir, 0755))
	 {
	    warn "Couldn't create $dir: $!\n";
	    $num_err++;
	 }
      }
   }
   return ($num_err == 0);
}



# ------------------------------ MNI Header ----------------------------------
#@NAME       : &check_output_path
#@INPUT      : $path   - path to a file or directory
#@OUTPUT     : 
#@RETURNS    : 0 if any component of $path is not a directory (or
#                cannot be created as one), or if last directory component
#                isn't a writeable directory (or can't be created).
#              1 otherwise
#@DESCRIPTION: Splits a file path up and makes sure that every
#              directory component except last is indeed a directory,
#              or can be created as one.  Makes sure that the last
#              component before a slash (last directory) is both a
#              directory and writeable; if not, attempts to create it.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1996/03/14, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub check_output_path
{
   my ($path) = @_;
   my (@dirs, $dir);

   # If $path starts with "/" (absolute path), then the first element
   # of @dirs will be blank, which we just strip off and ignore.  We
   # always strip off the last element of @dirs, because if $path ends
   # in a "/", then the last element is a blank and we don't care
   # about it (but we do check that the last component before the
   # slash is a writeable directory).  If $path didn't end with a "/",
   # then that indicates we're just making sure we can create the
   # named file, so we only look at the directory components (up to
   # the last slash).

   @dirs = split ("/", $path);
   if ($dirs[0] eq "")
   {
      $partial = "/";
      @dirs = @dirs[1..($#dirs-1)];
   }
   else
   {
      $partial = "";
      @dirs = @dirs[0..($#dirs-1)];
   }

   # Check all but the last directory component: each must be a directory.

   for $dir (@dirs[0..($#dirs-1)])
   {
#      my ($tmp) = "$partial$dir";

      $partial .= $dir;

      if (! -e $partial)
      {
	 unless (mkdir ($partial, 0755))
	 {
	    warn "Couldn't create $partial: $!\n";
	    return 0;
	 }
      }
      elsif (! -d $partial)
      {
	 warn "$path is not a writeable path because $partial is not a directory\n";
	 return 0;
      }
      $partial .= "/";
   }

   # Now check the last directory; must be a directory and writeable.

   $partial .= $dirs[-1];

   if (! -e $partial)
   {
      unless (mkdir ($partial, 0755))
      {
	 warn "Couldn't create $partial: $!\n";
	 return 0;
      }
   }
   elsif (! (-d $partial && -w $partial))
   {
      warn "$partial not a writeable directory\n";
      return 0;
   }

   return 1;
}
      



# ------------------------------ MNI Header ----------------------------------
#@NAME       : check_input_dirs
#@INPUT      : @dirs - list of directories to check
#@OUTPUT     : 
#@RETURNS    : 1 if all directories exist and are readable
#              0 otherwise
#@DESCRIPTION: Checks to see if all desired input directories exist and
#              are readable.  Prints meaningful error messages and returns
#              zero if anything is wrong.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/05/02, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub check_input_dirs
{
   my (@dirs) = @_;
   my ($num_err) = 0;

   foreach $dir (@dirs)
   {
      next unless $dir;		# skip blank strings
      if (-e $dir)
      {
	 if (! -d $dir)
	 {
	    warn "$dir exists but is not a directory\n";
	    $num_err++;
	 }
	 elsif (! -r $dir)
	 {
	    warn "$dir is a directory, but is not readable\n";
	    $num_err++;
	 }
      }
      else
      {
	 warn "directory $dir does not exist\n";
	 $num_err++;
      }
   }
   return ($num_err == 0);   
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : check_directories
#@INPUT      : $source - reference to list of source directories
#              $dest   - reference to list of destination directories
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Goes through the list of source directories, and makes sure
#              that each one exists; goes through list of destination
#              directories, and attempts to create those that don't exist.
#              Dies if any source directories are missing.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1994/08/22, Greg Ward
#@MODIFIED   : 1995/05/02, GW: changed to just call &check_input_dirs and
#                              &check_output_dirs
#-----------------------------------------------------------------------------
sub check_directories
{
   my ($source, $dest) = @_;
   my ($input_ok, $output_ok);

   $input_ok = &check_input_dirs (@$source);
   $output_ok = &check_output_dirs (@$dest);
   return ($input_ok && $output_ok);
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : search_directories
#@INPUT      : $file - name of file to search for
#              @dirs - list of directories to search in.  Note that
#                 the empty string should NOT be used; use "." for 
#                 the current directory.
#@OUTPUT     : 
#@RETURNS    : The directory where $file was found, or undef if it wasn't.
#@DESCRIPTION: Searches a list of directories for a single file.  Note
#              that only the existence test -e is used; doesn't check
#              for readability, executability, or whatever.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1994/09/16, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub search_directories
{
   my ($file, @dirs) = @_;
   my ($dir, $found);

   $found = 0;
   while ($dir = shift @dirs)
   {
      $dir .= "/" unless $dir =~ m|/$|;
      $found = (-e $dir . $file);
      last if $found;
   }

   return ($dir) if $found;
   return undef;
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : check_files
#@INPUT      : @files - a list of filenames to check for readability
#@OUTPUT     : 
#@RETURNS    : 1 if all files exist and are readable
#              0 otherwise (and prints a warning for every bad file)
#@DESCRIPTION: Makes sure that each of a list of files exists and is readable.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : Aug 1994, Greg Ward
#@MODIFIED   : 1995/04/25, GW: gave it two possible error messages
#              1995/10/30, GW: changed so it warns and returns 0 instead 
#                              of dying
#-----------------------------------------------------------------------------
sub check_files
{
   my (@files) = @_;
   my ($num_err);

   $num_err = 0;
   foreach (@files)
   {
      if (! -e $_)
      {	
	 warn "$_ does not exist\n";
	 $num_err++;
      }
      elsif (! -r $_)
      {
	 warn "$_ not readable\n";
	 $num_err++;
      }
   }
   return ($num_err == 0);
}



# ------------------------------ MNI Header ----------------------------------
#@NAME       : &find_program
#@INPUT      : $program - name of program to search for
#              $path    - colon-delimited list of directories to look in
#                         (if not supplied, $ENV{"PATH"} will be used);
#                         OR a reference to a list of directories
#@OUTPUT     : 
#@RETURNS    : full path of program, or 0 if not found (also prints a 
#              warning if not found)
#@DESCRIPTION: Searches a list of directories for an executable program.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/07/20, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub find_program
{
   my ($program, $path) = @_;
   my (@path, $found, $fullpath);

   $path = $ENV{'PATH'} unless defined $path;

   if (! ref $path)
   {
      @path = split (":", $path);
   }
   elsif (ref $path eq 'ARRAY')
   {
      @path = @$path;
   }
   else 
   {
      warn "\&find_program: \$path must be either a scalar or ref to array\n";
      return undef;
   }

   $found = 0;
   foreach $dir (@path)
   {
      $fullpath = "${dir}/${program}";
      if (-x $fullpath && -f $fullpath)
      {
	 $found = $fullpath;
	 last;
      }
   }

   return $found if $found;
   warn "Couldn't find program $program\n";
   return 0;
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &find_programs
#@INPUT      : @$programs  - list of programs to search for
#              $path       - reference to list of directories to search
#                            (OR a colon-delimited string)
#@OUTPUT     : 
#@RETURNS    : list of found programs (as full pathnames)
#              OR empty list (if there were any missing programs)
#@DESCRIPTION: Searches for each program specified in the specified search
#              path.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : &find_program
#@CREATED    : 1995/12/08, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub find_programs
{
   my ($programs, $path) = @_;
   my ($missing, $found, @found);

   $missing = 0;
   foreach $prog (@$programs)
   {
      $found = &find_program ($prog, $path);
      $missing++ if ! $found;
      push (@found, $found);
   }
   return @found if $missing == 0;
   return ();
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &uncompress
#@INPUT      : $tmp_dir - directory to uncompress files to
#              @originals - list of files to consider for uncompression
#@OUTPUT     : 
#@RETURNS    : array context: @originals, changed so that any the names
#                of any files that were compressed are now decompressed
#                and in $tmp_dir
#              scalar context: first element of @originals, possibly changed
#                to decompressed version
#@DESCRIPTION: Uncompresses (if applicable) compressed files to $tmp_dir.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/07/31, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub uncompress
{
   my ($tmp_dir, @originals) = @_;
   my ($uncompressed, @uncompressed);

   require "path_utilities.pl";	         # for &replace_dir

   foreach $orig (@originals)
   {
      if ($orig =~ /\.(Z|gz|z)$/)
      {
	 ($uncompressed = &replace_dir ($tmp_dir, $orig)) =~ s/\.(Z|gz|z)$//;
	 system ("gunzip -c $orig > $uncompressed");
	 die "gunzip $orig failed\n" if ($?);
	 push (@uncompressed, $uncompressed);
      }
      else
      {
	 push (@uncompressed, $orig);
      }
   }
   return @uncompressed if wantarray;
   return $uncompressed[0];
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &test_file
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : name of the file that ultimately passes the test
#@DESCRIPTION: Performs a single file test (eg. "-e" or "-r") with allowances
#              made for possible compressed files.  In particular, if the
#              the filename doesn't appear to be compressed already, then
#              we test $file, then $file.gz, $file.z, and $file.Z.  Only
#              if all of these tests fail does the test as a whole fail.
#              Returns the name of the file that ultimately passes the test.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : Greg Ward, March 1996
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub test_file
{
   my ($test, $file) = @_;

   # If the file as supplied passes the test, immediately return success
   
   return $file if (eval "$test \"$file\"");
   &Fatal ($@) if $@;

   # If the filename as supplied *is* compressed, then try doing the
   # test on the basename ($` is the portion of $file preceding the
   # match), and if successful return the matched filename.
   
   if ($file =~ /\.(Z|g?z)$/)
   {
      return $` if (eval "$test \"$`\"");
   }

   # If the filename as given is not compressed, then test the three
   # possible compressed formats -- return the matching filename as soon as
   # one of them matches.

   else
   {
      for $ending (qw(gz z Z))
      {
	 return "$file.$ending" if (eval "$test \"$file.$ending\"");
      }
   }
   
   # If we make it to here, then all tests failed -- return 0

   return 0;
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &generate_numbered_filename
#@INPUT      : $base - first part of filename
#@OUTPUT     : $ext  - last part of filename (including "." if wanted)
#@RETURNS    : Empty string on error (conflicting filenames or error 
#              renaming); otherwise, next filename in numbered sequence
#              starting with $base.
#@DESCRIPTION: Generates a numbered filename by incrementing a counter
#              $i until ${base}_${i}${ext} is found not to exist.  If
#              $i is 1 -- i.e. there weren't any files named with $base
#              and $ext -- then "_$i" is left out of the filename entirely.
#
#              For example, if called with $base="foo" and $ext=".log", and
#              neither "foo.log" nor "foo_1.log" exist, returns "foo.log".
#              On the next call, "foo.log" will be renamed to "foo_1.log",
#              and "foo_2.log" is returned.  Subsequent calls return
#              "foo_3.log", "foo_4.log", etc.  If both "foo.log" and
#              "foo_1.log" exist, then we print a warning and return the
#              empty string -- you'll have to deal with this degenerate
#              situation yourself, because it should never arise if you
#              only use &generate_numbered_filename to generate filenames.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1996/08/01, GPW, from code in ICBM.pm
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub generate_numbered_filename
{
   my ($base, $ext) = @_;
   my $i = 1;

   if (-e "${base}${ext}")
   {
      if (-e "${base}_1${ext}")
      {
	 warn "conflicting filenames: ${base}${ext} and ${base}_1${ext}";
         return '';
      }
      else
      {
	 rename ("${base}${ext}", "${base}_1${ext}") ||
            (warn "unable to rename ${base}${ext}: $!\n", return '');
      }
   }

   $i++ while (-e "${base}_${i}${ext}");
   ($i == 1) ? "${base}${ext}" : "${base}_${i}${ext}";
}


1;
