# ------------------------------ MNI Header ----------------------------------
#@NAME       : path_utilities.pl
#@DESCRIPTION: Functions for recognizing, parsing, and tweaking Unix
#              filenames and paths:
#                 &replace_dir
#                 &replace_ext
#                 &split_path
#@GLOBALS    : 
#@CREATED    : 1995/08/07, Greg Ward (from gpw_utilities.pl)
#@MODIFIED   : 
#@VERSION    : $Id: path_utilities.pl,v 1.7 1996/08/13 14:30:02 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------


sub ReplaceDir { &replace_dir; }
sub ReplaceExt { &replace_ext; }
sub SplitPath { &split_path; }
sub MergePaths { &merge_paths; }


# ------------------------------ MNI Header ----------------------------------
#@NAME       : replace_dir
#@INPUT      : $newpath - directory to replace existing directories with
#              @files   - list of files to have directories replaced
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Replaces the directory component of a list of pathnames.
#              Returns the list of files with substitutions performed.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/05/04, Greg Ward
#@MODIFIED   : 1995/05/23, GW: renamed to &ReplaceDir
#-----------------------------------------------------------------------------
sub replace_dir
{
   local ($newpath, @files) = @_;

   $newpath = "." if $newpath eq "";
   $newpath =~ s|([^/])$|$1/|;	# add a trailing slash if necessary
   foreach $file (@files)
   {
      $file =~ s|.*/||;		# strip off any existing directory
      $file = $newpath . $file;
   }
   return @files if wantarray;
   return $files[0];
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : replace_ext
#@INPUT      : $newext  - extension to replace existing extensions with
#              @files   - list of files to have extensions replaced
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Replaces the final extension (whatever follows the final dot)
#              of a list of pathnames.  Returns the list of files with
#              substitutions performed in array context, or the first filename
#              from the list in a scalar context.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/05/23, Greg Ward (from &ReplaceDir)
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub replace_ext
{
   local ($newext, @files) = @_;

   foreach $file (@files)
   {
      $file =~ s/\.[^\.]*/\.$newext/; # replace existing extension
   }
   return @files if wantarray;
   return $files[0];
}
      


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &split_path
#@INPUT      : $input - a Unix path specifiction (optional directory + file)
#              $ext_opt - specifies how to deal with file extensions:
#                         if "none", extension is ignored and returned as
#                           part of the base filename
#                         if "first", the *first* dot in a filename denotes
#                           the extension, eg ".mnc.gz" would be an extension
#                         if "last", the *last* dot denotes the extension,
#                           eg. just ".gz" would be the extension
#                         the default is "first"
#@OUTPUT     : 
#@RETURNS    : array: ($directory, $basename, $extension)
#@DESCRIPTION: Splits a Unix path specification into directory, base file 
#              name, and extension.  (The extension is chosen based on
#              either the first or last dot in the filename, depending
#              on the $ext_opt argument; by default, it splits on the 
#              first dot in the filename.)
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/05/10, Greg Ward - taken from mritotal and modified
#@MODIFIED   : 1995/08/10, GW: added $ext_opt option to handle splitting off
#                            the extension in different ways
#-----------------------------------------------------------------------------
sub split_path
{
    local($input, $ext_opt) = @_;
    local ($dir, $base, $ext);

    $ext_opt = "first" unless defined $ext_opt;

    # If filename has no extension, don't try to act as though it does
    # (both "last" and "first" options assume there is an extension

#    $ext_opt = "none" if $input !~ m+/?[^/]*\.+;

    if ($ext_opt eq "none")
    {
       ($dir, $base) = $input =~ m+^(.*/)?([^/]*)$+;
    } 
    elsif ($ext_opt eq "first")
    {
       ($dir, $base, $ext) = $input =~ m+^(.*/)?([^/\.]*)(\..*)?$+;
    }
    elsif ($ext_opt eq "last")
    {
       ($dir, $base, $ext) = $input =~ m+^(.*/)?([^/]*)(\.[^\/]*)$+;
    }
    else
    {
       die "split_path: unknown extension option \"$ext_opt\"\n";
    }

    if (!$dir)                  # no dir found? then use current dir
    {
       $dir = ".";
    }

    $dir =~ s/\/$//;		# strip off trailing slash

    ($dir, $base, $ext);
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &merge_paths
#@INPUT      : $path      - ref to a list of directories
#              $additions - ref to a list of directories to add to $path
#@OUTPUT     : @$path     - updated so elements from @$additions are added
#                           only if they weren't already in @$path
#@RETURNS    : 
#@DESCRIPTION: 
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 1995/12/04 GPW
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub merge_paths
{
   my ($path, $additions) = @_;

   foreach $add (@$additions)
   {
      push (@$path, $add) unless grep ($add eq $_, @$path);
   }
}



# ------------------------------ MNI Header ----------------------------------
#@NAME       : &expand_path
#@INPUT      : $path     - pathname to expand
#              $absolute - [flag; optional; default false]
#                          convert to absolute path? (not currently
#                          implemented -- how to do it???)
#              $env_vars - [flag; optional; default true] 
#                          expand $var to $ENV{'var'}?
#              $twiddles - [flag; optional; default true]
#                          expand ~user to `user''s home dir? (or 
#                          just ~ to $ENV{'HOME'})
#@OUTPUT     : 
#@RETURNS    : expanded version of $path
#@DESCRIPTION: Expands a pathname so that $VAR by the contents of the
#              environment variable VAR and ~username is replaced by the
#              username's login dir (or $ENV{'HOME'} if no username given).
#              Eventually, I'd like it to convert relative paths to
#              absolute, but I don't know an ironclad and efficient way to
#              do that.
#
#              Note: expansion of environment varibles is done first, so if
#              (for example) the environment variable FOO contains "~zap",
#              then "$FOO" will expand first to "~zap", and then to the
#              login directory of user zap.  This is different from the way
#              most shells behave.
#
#              Also, &expand_path will expand any number of occurences of
#              ~user or $VAR at any point in the string.  This, too, is
#              different from most shells.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : Aug 1996, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub expand_path
{
   my ($path, $absolute, $env_vars, $twiddles) = @_;
   my $home_dir_cache = {};

   $absolute = 0 unless defined $absolute;
   $env_vars = 1 unless defined $env_vars;
   $twiddles = 1 unless defined $twiddles;

   sub home_dir
   {
      my ($user) = @_;

      return $ENV{'HOME'} if $user eq "";
      return $home_dir_cache->{$user} if exists $home_dir_cache->{$user};
      $dir = (getpwnam ($user))[7];
      $home_dir_cache->{$user} = $dir;
   }

   die "expand_path: can't currently make pathnames absolute\n"
      if ($absolute);

   if ($env_vars)
   {
      1 while ($path =~ s|\$(\w+)|$ENV{$1}|e);
   }

   if ($twiddles)
   {
      1 while ($path =~ s|~(\w*)|&home_dir ($1)|e);
   }

   $path;
}

1;
